import torch
import torch.nn.functional as F

inputs = None
labels = None
model = None

def get_params_grad(model):
    params = []
    grads = []
    for _, param in model.named_parameters():
        if not param.requires_grad:
            continue
        params.append(param)
        grads.append(0. if param.grad is None else param.grad + 0.)
    return params, grads

def hessian_vector_product(gradsH, params, v):
    hv = torch.autograd.grad(
        gradsH, params, grad_outputs=v, only_inputs=True, retain_graph=True)
    return hv

def calculate_derivatives(epsilon_model):
    model.zero_grad()
    loss = F.cross_entropy(model(inputs), labels)
    loss.backward()
    params, grads = get_params_grad(model)
    epsilon_vec = [p.detach() for p in epsilon_model.parameters()]
    fd, sd = 0.0, 0.0
    fd = sum([(p.data * g.data).sum() for p, g in zip(epsilon_vec, grads)]).item()
    hv = hessian_vector_product(grads, params, v=epsilon_vec)
    sd = sum([(p.data * g.data).sum() for p, g in zip(epsilon_vec, hv)]).item()
    model.zero_grad()
    return fd, sd


lamb = 0.0
vs = []
for _ in range(100):
    epsilon_model = None  # sample Gaussian noise model
    fd, sd = calculate_derivatives(epsilon_model)
    vs.append(lamb * fd + 0.5 * lamb * lamb * sd)