import torch
import torch.nn.functional as F
from torch.autograd import grad
from scipy.sparse.linalg import LinearOperator, eigsh

inputs = None
labels = None
model = None
n_params = None  # number of total trainable parameters

def hessian_vector_product(vector):
    loss = F.cross_entropy(model(inputs), labels)
    grad_params = grad(loss, model.parameters(), create_graph=True)
    flat_grad = torch.cat([g.view(-1) for g in grad_params])
    grad_vector_product = torch.sum(flat_grad * vector)
    hvp = grad(grad_vector_product, model.parameters(), retain_graph=True)
    return torch.cat([g.contiguous().view(-1) for g in hvp])

def matvec(v):
    v_tensor = torch.tensor(v, dtype=torch.float32)
    return hessian_vector_product(v_tensor).cpu().detach().numpy()

eigenvalues, eigenvectors = eigsh(
    LinearOperator((n_params, n_params), matvec=matvec),
    k=10, tol=1e-3, which="SA", return_eigenvectors=True
)